/*
 * Raytracing v3.0
 *
 * Compile and build:
 * zcc +cpm -subtype=pcw80 -compiler=sccz80 -O2 -lm -create-app raytr.c pbmout.c kernel.c screen.c buffered_io.c readscene.c -o raytr.com
 */
#include <stdio.h>
#include <conio.h>
#include <ctype.h>
#include <graphics.h>
#include <math.h>
#include "kernel.h"
#include "mytypes.h"
#include "pbmout.h"
#include "readscene.h"

#define CH_BEEP 7
#define XM 4
#define YM 4
#define SPHERE_COUNT 4
#define XRES 720
#define YRES 256
#define XRES_HALF 360
#define YRES_HALF 128
#define XMAX 719
#define YMAX 255

unsigned char floorColour[] = {0, 3, 6, 11}; // colours to plot the floor
t_scene *myscene;
float q[4];
float x, y, z;
float dx, dy, dz;

// function prototypes
void trace(int i, int j);
float xMinusIntX(float f);
void plotGrey(int x, int y, unsigned char r);
void beep(void);
void invertInMemory(void);
char *getFilename(int argc, char *argv[], const char *option);

/*
 * main
 */
int main(int argc, char *argv[])
{
    int i, j;
	char ch;
	char *pbmfilename;
	char *scenefilename;
	float r;
    
	gx_init();
    clg();
    printf("%cH", 27); // ESC H = VT100 cursor home
    printf("%c0", 27); // ESC 0 = switch off status bar

	// evaluate commandline args
	pbmfilename   = getFilename(argc, argv, "-O"); // output file
	scenefilename = getFilename(argc, argv, "-I"); // input file
	// set default values
	if (pbmfilename == NULL) {
		pbmfilename = "SCENE.PBM";
	}
	if (scenefilename == NULL) {
		scenefilename = "SCENE.TXT";
	}
	
	myscene = readScene(scenefilename);

	// set radius squares
    for (i = 0; i < myscene->sphereCount; ++i) {
		r = myscene->radius[i];
        q[i] = r * r;
    }
    
	// trace the scene
    for (i = 0; i < YRES * 2; i += YM) {
        for (j = 0; j < XRES; j += XM) {
            x = myscene->camera.x;
            y = myscene->camera.y;
            z = myscene->camera.z;
            dx = (float) (j - XRES_HALF);
            dy = (float) (YRES - i);
            dz = 450.0;
            trace(i, j);
        }
    }

    beep();
	
	// wait for user input: I = invert, W = write, Q = quit, X = exit
	do {
		ch = getch();
		ch = toupper(ch);
		switch (ch) {
			case 'I': beep();
			          invertInMemory();
					  beep();
			          break;
			case 'W': beep();
			          saveScreenToPBM(pbmfilename, true);
					  beep();
			          break;
		}
	} while (ch != 'Q' && ch != 'X');
	
	beep();
    clg();
    printf("%c1", 27); // ESC 1 = switch on status bar
    printf("%ce", 27); // ESC e = show cursor
    
    return 0;
}

/*
 * Traces a primary ray.
 */
void trace(int i, int j)
{
    float dd;
    float s;
    float px, py, pz;
    float pp;
    float sc;
    float aa, bb;
    char n;
    unsigned char k, kn;
    unsigned char r1;
    unsigned char ba;
    float r1f;
    float nx, ny, nz;
    float nn;
    float l;
    float u, v;
	t_coords3d *coords;
    
    dd = dx * dx + dy * dy + dz * dz;
    s = 0.0;
    ba = 2;
    
    for (;;) {
        if (y >= 0 || dy <= 0) {
            n = -1;
        }
        else {
            n = 0;
        }
		
        if (!n) {
            s = -y / dy;
        }
		
        for (k = 0; k < SPHERE_COUNT; ++k) {
			coords = &myscene->sphere[k];
			px = coords->x - x;
			py = coords->y - y;
			pz = coords->z - z;
            pp = px * px + py * py + pz * pz;
            sc = px * dx + py * dy + pz * dz;
			
            if (sc > 0) {
                bb = sc * sc / dd;
                aa = q[k] - pp + bb;
                if (aa > 0) {
                    sc = (sqrt(bb) - sqrt(aa)) / sqrt(dd);
                    if (sc < s || n < 0) {
                        kn = k;
						n = 1;
                        s = sc;
                    }
                }
            }
        }
		
        if (n < 0) {
            r1f = dy * dy / dd * 11.0;
            r1 = (unsigned char) r1f;
            ++r1;
            plotGrey(j, i, r1);
            n = 0;
            return;
        }
		
        dx *= s;
        dy *= s;
        dz *= s;
        dd *= s * s;
        x += dx;
        y += dy;
        z += dz;
		
        if (n == 0) {
            break;
        }
		
		coords = &myscene->sphere[kn];
		nx = x - coords->x;
		ny = y - coords->y;
		nz = z - coords->z;
        nn = nx * nx + ny * ny + nz * nz;
        l = 2 * (dx * nx + dy * ny + dz * nz) / nn;
        dx -= nx * l;
        dy -= ny * l;
        dz -= nz * l;
    }

    for (k = 0; k < SPHERE_COUNT; ++k) {
		coords = &myscene->sphere[k];
        u = coords->x - x;
        v = coords->z - z;
        if (u * u + v * v <= q[k]) {
            ba = 0;
        }
    }
    
    if ((xMinusIntX(x) > 0.5) == (xMinusIntX(z) > 0.5)) {
        r1 = floorColour[ba];
    }
    else {
        r1 = floorColour[ba + 1];
    }
	
    plotGrey(j, i, r1);
}

float xMinusIntX(float f)
{
    int i;

	if (f < 0) {
		f = f + 0.000001; // avoid rounding bug
		f = -f;
		i = (int) f;
		return 1.0 - f + (float) i;
	}
	else {
		i = (int) f;
		return f - (float) i;
	}
}

/*
 * Plots a "grey" macro pixel built from 4x2 regular pixels.
 */
void plotGrey(int x, int y, unsigned char greyValue)
{
    y = YMAX - (y >> 1);
	
    switch(greyValue) {
        case  0:
            break;
        case  1:
            plot(++x, y);
            break;
        case  2:
            plot(++x, y);
            plot(++x, ++y);
            break;
        case  3:
            plot(x++, y);
            plot(++x, y);
            plot(--x, ++y);
            break;
        case  4:
            plot(x++, y);
            plot(++x, y);
            plot(--x, ++y);
            ++x;
            plot(++x, y);
            break;
        case  5:
            plot(x++, y);
            ++x;
            plot(++x, y);
            --x;
            plot(--x, ++y);
            plot(++x, y);
            break;
        case  6:
            plot(x, y);
            plot(++x, y);
            plot(++x, y);
            plot(--x, ++y);
            ++x;
            plot(++x, y);
            break;
        case  7:
            plot(x, y);
            plot(++x, y);
            ++x;
            plot(++x, y);
            --x;
            plot(--x, ++y);
            plot(++x, y);
            break;
        case  8:
            plot(x, y);
            plot(++x, y);
            plot(++x, y);
            plot(++x, y);
            --x;
            plot(--x, ++y);
            plot(++x, y);
            break;
        case  9:
            plot(x, y);
            plot(++x, y);
            plot(++x, y);
            plot(--x, ++y);
            plot(++x, y);
            plot(++x, y);
            break;
        case 10:
            plot(x, y);
            plot(++x, y);
            plot(++x, y);
            plot(++x, y);
            plot(--x, ++y);
            plot(--x, y);
            plot(--x, y);
            break;
        case 11:
            plot(x, y);
            plot(++x, y);
            plot(++x, y);
            plot(++x, y);
            plot(x, ++y);
            plot(--x, y);
            plot(--x, y);
            plot(--x, y);
            break;
        default:
            printf("%d ", greyValue);
    }
}

/*
 * Emits a beep through the loudspeaker.
 */
void beep()
{
    putchar(CH_BEEP);
}

/*
 * Gets the filename for the option provided from the commandline.
 */
char *getFilename(int argc, char *argv[], const char *option) {
	
	int i;
	char *filename;
	char optionFound;
	
	filename = NULL;
	optionFound = 0;
	
	for (i = 1; i < argc; ++i) {
		if (optionFound) {
			filename = argv[i];
			break;
		}
		
		if (strcmp(argv[i], option) == 0) {
			optionFound = 1;
		}
	}
	
	return filename;
}
